/* Copyright 2007 The JA-SIG Collaborative.  All rights reserved.
*  See license distributed with this file and
*  available online at http://www.uportal.org/license.html
*/

package org.jasig.portal.tools.deployer.tomcat;

import java.io.File;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.util.jar.JarEntry;
import java.util.jar.JarFile;

import org.apache.commons.io.FileUtils;
import org.apache.commons.io.IOUtils;
import org.jasig.portal.tools.deployer.AbstractEarDeployer;
import org.jasig.portal.tools.deployer.DeployerConfig;
import org.jasig.portal.tools.deployer.WebModule;

/**
 * Encapsulates Tomcat specific logic for deploying WARs and JARs from the uPortal3 EAR.
 * 
 * @author Eric Dalquist
 * @version $Revision$
 */
public class TomcatEarDeployer extends AbstractEarDeployer {
    
    /**
     * Writes the WAR to Tomcat's webapps directory, as specified by {@link TomcatDeployerConfig#getCatalinaWebapps()}.
     * 
     * @see org.jasig.portal.tools.deployer.AbstractEarDeployer#deployWar(org.jasig.portal.tools.deployer.WebModule, java.util.jar.JarFile, org.jasig.portal.tools.deployer.DeployerConfig)
     */
    @Override
    protected final void deployWar(WebModule webModule, JarFile earFile, DeployerConfig deployerConfig) throws IOException {
        final TomcatDeployerConfig tomcatDeployerConfig = this.getTomcatDeployerConfig(deployerConfig);
        
        final String webUri = webModule.getWebUri();
        final JarEntry warEntry = earFile.getJarEntry(webUri);

        String contextRoot = webModule.getContextRoot();
        if (!contextRoot.endsWith(".war")) {
            contextRoot += ".war";
        }
        if (contextRoot.startsWith("/")) {
            contextRoot = contextRoot.substring(1);
        }
        final File warDest = this.createSafeFile(tomcatDeployerConfig.getCatalinaWebapps(), contextRoot);
        
        this.copyAndClose(warEntry, earFile, warDest);
    }

    /**
     * Writes the JAR to Tomcat's shared/lib directory, as specified by {@link TomcatDeployerConfig#getCatalinaShared()}.
     * 
     * @see org.jasig.portal.tools.deployer.AbstractEarDeployer#deployJar(java.util.jar.JarEntry, java.util.jar.JarFile, org.jasig.portal.tools.deployer.DeployerConfig)
     */
    @Override
    protected final void deployJar(JarEntry jarEntry, JarFile earFile, DeployerConfig deployerConfig) throws IOException {
        final TomcatDeployerConfig tomcatDeployerConfig = this.getTomcatDeployerConfig(deployerConfig);
        
        final String jarName = jarEntry.getName();
        final File sharedLib = new File(tomcatDeployerConfig.getCatalinaShared(), "lib");
        final File jarDest = this.createSafeFile(sharedLib, jarName);
        
        this.copyAndClose(jarEntry, earFile, jarDest);
    }

    /**
     * Does a type check and converts a generic {@link DeployerConfig} to a {@link TomcatDeployerConfig}.
     * 
     * @param deployerConfig The config to try and cast
     * @return A TomcatDeployerConfig.
     */
    protected TomcatDeployerConfig getTomcatDeployerConfig(DeployerConfig deployerConfig) {
        if (!(deployerConfig instanceof TomcatDeployerConfig)) {
            throw new IllegalArgumentException("deployerConfig must be of type " + TomcatDeployerConfig.class);
        }
        final TomcatDeployerConfig tomcatDeployerConfig = (TomcatDeployerConfig)deployerConfig;
        return tomcatDeployerConfig;
    }

    /**
     * Creates a File for the specified directory and name, if a file already exists
     * at the location it is deleted and all parent directories are verfied to exist. 
     * 
     * @param baseDir The directory to base the file in
     * @param fileName The name for the file
     * @return A new File object that has its parent directories and an existing file deleted. 
     */
    protected File createSafeFile(final File baseDir, final String fileName) throws IOException {
        final File safeFile = new File(baseDir, fileName);
        if (safeFile.exists()) {
            safeFile.delete();
        }
        else {
            FileUtils.forceMkdir(baseDir);
        }
        
        return safeFile;
    }

    /**
     * Reads the specified {@link JarEntry} from the {@link JarFile} and writes its contents
     * to the specified {@link File}.
     * 
     * @param jarEntry The JarEntry for the file to read from the archive.
     * @param jarFile The JarFile to get the {@link InputStream} for the file from.
     * @param destinationFile The File to write to, all parent directories should exist and no file should already exist at this location.
     * @throws IOException If the copying of data from the JarEntry to the File fails.
     */
    protected void copyAndClose(JarEntry jarEntry, JarFile jarFile, File destinationFile) throws IOException {
        if (this.logger.isInfoEnabled()) {
            this.logger.info("Copying EAR entry '" + jarFile.getName() + "!" + jarEntry.getName() + "' to '" + destinationFile + "'");
        }

        final InputStream jarEntryStream = jarFile.getInputStream(jarEntry);
        try {
            final OutputStream jarOutStream = new FileOutputStream(destinationFile);
            try {
                IOUtils.copy(jarEntryStream, jarOutStream);
            }
            finally {
                IOUtils.closeQuietly(jarOutStream);
            }
        }
        finally {
            IOUtils.closeQuietly(jarEntryStream);
        }
    }
}
